@@ -1,19 +1,19 @@
 #!/usr/bin/python
 #
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
-# 
+#
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
 # Free Software Foundation; either version 2 of the License, or (at your
 # option) any later version.
-# 
+#
 # This program is distributed in the hope that it will be useful, but
 # WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 # Public License for more details.
-# 
+#
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc., 
+# this program; if not, write to the Free Software Foundation, Inc.,
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 
 
@@ -28,8 +28,10 @@ import sys
 import stat
 import time
 import logging
-from signal import alarm, signal, SIGALRM, SIGKILL
+from signal import alarm, signal, SIGALRM, SIGINT, SIGKILL
 import subprocess
+import fcntl
+import random
 import ConfigParser
 
 import piupartslib.conf
@@ -39,6 +41,9 @@ import piupartslib.packagesdb
 CONFIG_FILE = "/etc/piuparts/piuparts.conf"
 MAX_WAIT_TEST_RUN = 45*60
 
+interrupted = False
+old_sigint_handler = None
+
 def setup_logging(log_level, log_file_name):
     logger = logging.getLogger()
     logger.setLevel(log_level)
@@ -57,7 +62,7 @@ def setup_logging(log_level, log_file_name):
 
 class Config(piupartslib.conf.Config):
 
-    def __init__(self, section="slave"):
+    def __init__(self, section="slave", defaults_section=None):
         self.section = section
         piupartslib.conf.Config.__init__(self, section,
             {
@@ -74,6 +79,7 @@ class Config(piupartslib.conf.Config):
                 "mirror": None,
                 "piuparts-cmd": "sudo piuparts",
                 "distro": None,
+                "area": None,
                 "chroot-tgz": None,
                 "upgrade-test-distros": None,
                 "upgrade-test-chroot-tgz": None,
@@ -82,7 +88,8 @@ class Config(piupartslib.conf.Config):
                 "keep-sources-list": "no",
                 "arch": None,
                 "precedence": "1",
-            }, "")
+            },
+            defaults_section=defaults_section)
 
 
 class Alarm(Exception):
@@ -91,6 +98,20 @@ class Alarm(Exception):
 def alarm_handler(signum, frame):
     raise Alarm
 
+def sigint_handler(signum, frame):
+    global interrupted
+    interrupted = True
+    print '\nSlave interrupted by the user, waiting for the current test to finish.'
+    print 'Press Ctrl-C again to abort now.'
+    signal(SIGINT, old_sigint_handler)
+
+
+class MasterIsBusy(Exception):
+
+    def __init__(self):
+        self.args = "Master is busy, retry later"
+
+
 class MasterNotOK(Exception):
 
     def __init__(self):
@@ -153,13 +174,15 @@ class Slave:
         else:
             user = ""
         ssh_cmdline = "cd %s; %s 2> %s.$$ && rm %s.$$" % \
-                      (self._master_directory or ".", 
+                      (self._master_directory or ".",
                       self._master_command, log_file, log_file)
-        p = subprocess.Popen(["ssh", user + self._master_host, ssh_cmdline], 
+        p = subprocess.Popen(["ssh", user + self._master_host, ssh_cmdline],
                        stdin=subprocess.PIPE, stdout=subprocess.PIPE)
         self._to_master = p.stdin
         self._from_master = p.stdout
         line = self._readline()
+        if line == "busy\n":
+            raise MasterIsBusy()
         if line != "hello\n":
             raise MasterDidNotGreet()
         logging.debug("Connected to master")
@@ -235,68 +258,111 @@ class Slave:
 class Section:
 
     def __init__(self, section):
-        self._config = Config(section=section)
+        self._config = Config(section=section, defaults_section="global")
         self._config.read(CONFIG_FILE)
+        self._error_wait_until = 0
+        self._idle_wait_until = 0
         self._slave_directory = os.path.abspath(self._config["slave-directory"])
         if not os.path.exists(self._slave_directory):
-            os.mkdir(self._slave_directory)
+            os.makedirs(self._slave_directory)
 
-    def setup(self, master_host, master_user, master_directory, base_tgz_ctrl):
         if self._config["debug"] in ["yes", "true"]:
             self._logger = logging.getLogger()
             self._logger.setLevel(logging.DEBUG)
 
-        oldcwd = os.getcwd()
-        os.chdir(self._slave_directory)
-
         if self._config["chroot-tgz"] and not self._config["distro"]:
           logging.info("The option --chroot-tgz needs --distro.")
 
-        self._base_tgz_ctrl = base_tgz_ctrl
-        self._check_tarball()
+        if int(self._config["max-reserved"]) > 0:
+            self._check_tarball()
 
-        for rdir in ["new", "pass", "fail"]:
+        for rdir in ["new", "pass", "fail", "untestable", "reserved"]:
             rdir = os.path.join(self._slave_directory, rdir)
             if not os.path.exists(rdir):
                 os.mkdir(rdir)
 
         self._slave = Slave()
-        self._slave.set_master_host(master_host)
-        self._slave.set_master_user(master_user)
-        self._slave.set_master_directory(master_directory)
-        self._slave.set_master_command(self._config["master-command"])
-        self._log_file=self._config["log-file"]
 
-        for rdir in ["pass", "fail", "untestable", "reserved"]:
-            rdir = os.path.join(self._slave_directory, rdir)
-            if not os.path.exists(rdir):
-                os.makedirs(rdir)
-        os.chdir(oldcwd)
+    def _connect_to_master(self):
+        self._slave.set_master_host(self._config["master-host"])
+        self._slave.set_master_user(self._config["master-user"])
+        self._slave.set_master_directory(self._config["master-directory"])
+        self._slave.set_master_command(self._config["master-command"] + " " + self._config.section)
+        self._slave.connect_to_master(self._config["log-file"])
 
     def _check_tarball(self):
+        oldcwd = os.getcwd()
+        os.chdir(self._slave_directory)
+
         tarball = self._config["chroot-tgz"]
         if tarball:
             create_or_replace_chroot_tgz(self._config, tarball,
-                      self._base_tgz_ctrl, self._config["distro"])
+                                         self._config["distro"])
 
         tarball = self._config["upgrade-test-chroot-tgz"]
         if self._config["upgrade-test-distros"] and tarball:
             create_or_replace_chroot_tgz(self._config, tarball,
-                      self._base_tgz_ctrl, self._config["upgrade-test-distros"].split()[0])
+                                         self._config["upgrade-test-distros"].split()[0])
+
+        os.chdir(oldcwd)
 
     def precedence(self):
         return int(self._config["precedence"])
 
+    def sleep_until(self):
+        return max(self._error_wait_until, self._idle_wait_until)
+
     def run(self):
+        if time.time() < self.sleep_until():
+            return 0
+
         logging.info("-------------------------------------------")
         logging.info("Running section %s (precedence=%d)" % (self._config.section, self.precedence()))
-        self._config = Config(section=self._config.section)
+        self._config = Config(section=self._config.section, defaults_section="global")
         self._config.read(CONFIG_FILE)
-        self._slave.connect_to_master(self._log_file)
+
+        if int(self._config["max-reserved"]) == 0:
+            logging.info("disabled")
+            self._error_wait_until = time.time() + 12 * 3600
+            return 0
+
+        if not self._config["distro"] and (not self._config["upgrade-test-distros"] \
+                                           or not self._config["upgrade-test-distros"].split()):
+            logging.error("neither 'distro' nor 'upgrade-test-distros' configured")
+            self._error_wait_until = time.time() + 3600
+            return 0
+
+        lock = open(os.path.join(self._slave_directory, "slave.lock"), "we")
+        try:
+            fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
+        except IOError:
+            logging.info("busy")
+            self._error_wait_until = time.time() + 900
+            lock.close()
+            return 0
 
         oldcwd = os.getcwd()
         os.chdir(self._slave_directory)
-        test_count = 0
+
+        ret = self._run()
+
+        os.chdir(oldcwd)
+        lock.close()
+        return ret
+
+    def _run(self):
+        try:
+            self._connect_to_master()
+        except KeyboardInterrupt:
+            raise
+        except MasterIsBusy:
+            logging.error("master is busy")
+            self._error_wait_until = time.time() + random.randrange(60, 180)
+            return 0
+        except:
+            logging.error("connection to master failed")
+            self._error_wait_until = time.time() + 900
+            return 0
 
         for logdir in ["pass", "fail", "untestable"]:
             for basename in os.listdir(logdir):
@@ -313,42 +379,54 @@ class Section:
         self._slave.get_status(self._config.section)
         self._slave.close()
 
-        if self._slave.get_reserved():
-            self._check_tarball()
-            packages_files = {}
-            if self._config["distro"]:
-                distros = [self._config["distro"]]
-            else:
-                distros = []
+        if not self._slave.get_reserved():
+            self._idle_wait_until = time.time() + int(self._config["idle-sleep"])
+            return 0
 
-            if self._config["upgrade-test-distros"]:
-                distros += self._config["upgrade-test-distros"].split()
+        if self._config["distro"]:
+            distros = [self._config["distro"]]
+        else:
+            distros = []
 
-            for distro in distros:
-                if distro not in packages_files:
+        if self._config["upgrade-test-distros"]:
+            distros += self._config["upgrade-test-distros"].split()
+
+        packages_files = {}
+        for distro in distros:
+            if distro not in packages_files:
+                try:
                     packages_files[distro] = fetch_packages_file(self._config, distro)
-            if self._config["distro"]:
-              packages_file = packages_files[self._config["distro"]]
-            else:
-              packages_file = packages_files[distro]
+                except IOError:
+                    logging.error("failed to fetch packages file for %s" % distro)
+                    self._error_wait_until = time.time() + 900
+                    return 0
+        if self._config["distro"]:
+            packages_file = packages_files[self._config["distro"]]
+        else:
+            packages_file = packages_files[distro]
 
-            for package_name, version in self._slave.get_reserved():
-                test_count += 1
-                if package_name in packages_file:
-                    package = packages_file[package_name]
-                    if version == package["Version"] or self._config["upgrade-test-distros"]:
-                        test_package(self._config, package, packages_files)
-                    else:
-                        create_file(os.path.join("untestable", 
-                                    log_name(package_name, version)),
-                                    "%s %s not found" % (package_name, version))
+        test_count = 0
+        self._check_tarball()
+        for package_name, version in self._slave.get_reserved():
+            test_count += 1
+            if package_name in packages_file:
+                package = packages_file[package_name]
+                if version == package["Version"]:
+                    test_package(self._config, package, packages_files)
                 else:
-                    create_file(os.path.join("untestable", 
+                    logging.info("Cannot test %s/%s %s" % (self._config.section, package_name, version))
+                    create_file(os.path.join("untestable",
                                 log_name(package_name, version)),
-                                "Package %s not found" % package_name)
-                self._slave.forget_reserved(package_name, version)
-
-        os.chdir(oldcwd)
+                                "%s %s not found, %s is available\n" \
+                                    % (package_name, version, package["Version"]))
+            else:
+                logging.info("Cannot test %s/%s %s" % (self._config.section, package_name, version))
+                create_file(os.path.join("untestable",
+                            log_name(package_name, version)),
+                            "Package %s not found\n" % package_name)
+            self._slave.forget_reserved(package_name, version)
+            if interrupted:
+                raise KeyboardInterrupt
         return test_count
 
 
@@ -369,92 +447,123 @@ def upgrade_testable(config, package, packages_files):
     else:
         return False
 
-def get_process_children(pid):
-    p = subprocess.Popen('ps --no-headers -o pid --ppid %d' % pid,
-           shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
-    stdout, stderr = p.communicate()
-    return [int(p) for p in stdout.split()]
 
-def run_test_with_timeout(cmd, maxwait, kill_all):
-      logging.debug("Executing: %s" % cmd)
+def run_test_with_timeout(cmd, maxwait, kill_all=True):
 
-      stdout = ""
-      sh_cmd = "{ %s; } 2>&1" % cmd
-      p = subprocess.Popen(sh_cmd, shell=True,
-                     stdin=subprocess.PIPE, stdout=subprocess.PIPE)
-      if maxwait > 0:
-          signal(SIGALRM, alarm_handler)
-          alarm(maxwait)
-      try:
-          stdout, stderr = p.communicate()
-          if maxwait > 0:
-              alarm(0)
-      except Alarm:
-          pids = [p.pid]
-          if kill_all:
-              pids.extend(get_process_children(p.pid))
-          for pid in pids:
-              if pid > 0:
-                  try: 
-                      os.kill(pid, SIGKILL)
-                  except OSError:
-                      pass
-          return -1,stdout
+    def terminate_subprocess(p, kill_all):
+        pids = [p.pid]
+        if kill_all:
+            ps = subprocess.Popen(["ps", "--no-headers", "-o", "pid", "--ppid", "%d" % p.pid],
+                                  stdout = subprocess.PIPE)
+            stdout, stderr = ps.communicate()
+            pids.extend([int(pid) for pid in stdout.split()])
+        if p.poll() is None:
+            print 'Sending SIGINT...'
+            try:
+                os.killpg(os.getpgid(p.pid), SIGINT)
+            except OSError:
+                pass
+            # piuparts has 30 seconds to clean up after Ctrl-C
+            for i in range(60):
+                time.sleep(0.5)
+                if p.poll() is not None:
+                    break
+        if p.poll() is None:
+            print 'Sending SIGTERM...'
+            p.terminate()
+            # piuparts has 5 seconds to clean up after SIGTERM
+            for i in range(10):
+                time.sleep(0.5)
+                if p.poll() is not None:
+                    break
+        if p.poll() is None:
+            print 'Sending SIGKILL...'
+            p.kill()
+        for pid in pids:
+            if pid > 0:
+                try:
+                    os.kill(pid, SIGKILL)
+                    print "Killed %d" % pid
+                except OSError:
+                    pass
 
-      return p.returncode,stdout 
+    logging.debug("Executing: %s" % " ".join(cmd))
+
+    stdout = ""
+    p = subprocess.Popen(cmd, preexec_fn=os.setpgrp,
+                         stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+    if maxwait > 0:
+        signal(SIGALRM, alarm_handler)
+        alarm(maxwait)
+    try:
+        stdout, stderr = p.communicate()
+        alarm(0)
+    except Alarm:
+        terminate_subprocess(p, kill_all)
+        return -1,stdout
+    except KeyboardInterrupt:
+        print '\nSlave interrupted by the user, cleaning up...'
+        try:
+            kill_subprocess(p, kill_all)
+        except KeyboardInterrupt:
+            print '\nTerminating piuparts was interrupted... manual cleanup still neccessary.'
+            raise
+        raise
+
+    return p.returncode,stdout
 
 
 def test_package(config, package, packages_files):
+    global old_sigint_handler
+    old_sigint_handler = signal(SIGINT, sigint_handler)
     logging.info("Testing package %s/%s %s" % (config.section, package["Package"], package["Version"]))
 
     output_name = log_name(package["Package"], package["Version"])
     logging.debug("Opening log file %s" % output_name)
     new_name = os.path.join("new", output_name)
-    output = file(new_name, "w")
-    output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n", 
+    output = file(new_name, "we")
+    output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n",
                                time.gmtime()))
     output.write("\n")
     package.dump(output)
     output.write("\n")
 
-    # omit distro test if chroot-tgz is not specified.
-    ret = 0
-    if config["chroot-tgz"]: 
-        command = "%(piuparts-cmd)s -ad %(distro)s -b %(chroot-tgz)s" % \
-                    config
-        if config["keep-sources-list"] in ["yes", "true"]:
-            command += " --keep-sources-list "
-        if config["mirror"]:
-            command += " --mirror %s " % config["mirror"]
-        command += " " + package["Package"]
+    base_command = config["piuparts-cmd"].split()
+    if config["mirror"]:
+        base_command.extend(["--mirror", config["mirror"]])
 
-        output.write("Executing: %s\n" % command)
-        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN, True)
+    ret = 0
+    if config["chroot-tgz"]:
+        command = base_command[:]
+        command.extend(["-b", config["chroot-tgz"]])
+        command.extend(["-d", config["distro"]])
+        if config["keep-sources-list"] in ["yes", "true"]:
+            command.append("--keep-sources-list")
+        command.extend(["--apt", "%s=%s" % (package["Package"], package["Version"])])
+
+        output.write("Executing: %s\n" % " ".join(command))
+        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
         if ret < 0:
             output.write(f + "\n *** Process KILLED - exceed maximum run time ***\n")
         else:
             output.write(f)
 
-    if ret == 0 and upgrade_testable(config, package, packages_files):
-        distros = config["upgrade-test-distros"].split()
-        distros = ["-d " + distro.strip() for distro in distros]
-        distros = " ".join(distros)
-        command = "%(piuparts-cmd)s -ab %(upgrade-test-chroot-tgz)s " % config
-        command += distros
-        if config["mirror"]:
-          command += " --mirror %s " % config["mirror"]
-        command += " " + package["Package"]
+    if ret == 0 and config["upgrade-test-chroot-tgz"] and upgrade_testable(config, package, packages_files):
+        command = base_command[:]
+        command.extend(["-b", config["upgrade-test-chroot-tgz"]])
+        for distro in config["upgrade-test-distros"].split():
+            command.extend(["-d", distro])
+        command.extend(["--apt", "%s=%s" % (package["Package"], package["Version"])])
 
-        output.write("Executing: %s\n" % command)
-        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN, True)
+        output.write("Executing: %s\n" % " ".join(command))
+        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
         if ret < 0:
             output.write(" *** Process KILLED - exceed maximum run time ***\n")
         else:
             output.write(f)
-            output.flush()
 
     output.write("\n")
-    output.write(time.strftime("End: %Y-%m-%d %H:%M:%S %Z\n", 
+    output.write(time.strftime("End: %Y-%m-%d %H:%M:%S %Z\n",
                                time.gmtime()))
     output.close()
     if ret != 0:
@@ -462,14 +571,15 @@ def test_package(config, package, packages_files):
     else:
         subdir = "pass"
     os.rename(new_name, os.path.join(subdir, output_name))
-    logging.debug("Done with %s" % output_name)
+    logging.debug("Done with %s: %s (%d)" % (output_name, subdir, ret))
+    signal(SIGINT, old_sigint_handler)
 
 
 def create_chroot(config, tarball, distro):
     output_name = tarball + ".log"
     logging.debug("Opening log file %s" % output_name)
     logging.info("Creating new tarball %s" % tarball)
-    command = "%s -ad %s -s %s.new -m %s hello" % \
+    command = "%s -ad %s -s %s.new -m %s dpkg" % \
                 (config["piuparts-cmd"], distro, tarball, config["mirror"])
     output = file(output_name, "w")
     output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
@@ -487,11 +597,11 @@ def create_chroot(config, tarball, distro):
     if os.path.exists(tarball + ".new"):
         os.rename(tarball + ".new", tarball)
 
-def create_or_replace_chroot_tgz(config, tgz, tgz_ctrl, distro):
-    forced = 0 
+def create_or_replace_chroot_tgz(config, tgz, distro):
+    forced = 0
     if os.path.exists(tgz):
-        max_tgz_age = tgz_ctrl[0]
-        min_tgz_retry_delay = tgz_ctrl[1]
+        max_tgz_age = int(config["max-tgz-age"])
+        min_tgz_retry_delay = int(config["min-tgz-retry-delay"])
         now = time.time()
         statobj = os.stat(tgz)
         # stat.ST_MTIME is actually time file was initially created
@@ -502,7 +612,7 @@ def create_or_replace_chroot_tgz(config, tgz, tgz_ctrl, distro):
             # stat.ST_CTIME is time created OR last renamed
             if min_tgz_retry_delay is None or now - statobj[stat.ST_CTIME] > min_tgz_retry_delay:
                 os.rename(tgz, tgz + ".old")
-                forced = 1 
+                forced = 1
                 logging.info("%s too old.  Renaming to force re-creation" % tgz)
     if not os.path.exists(tgz):
         create_chroot(config, tgz, distro)
@@ -514,16 +624,7 @@ def create_or_replace_chroot_tgz(config, tgz, tgz_ctrl, distro):
                 os.unlink(tgz + ".old")
 
 def fetch_packages_file(config, distro):
-    mirror = config["mirror"]
-    arch = config["arch"]
-    if not arch:
-        # Try to figure it out ourselves, using dpkg
-        p = subprocess.Popen(["dpkg", "--print-architecture"], 
-                             stdout=subprocess.PIPE)
-        arch = p.stdout.read().rstrip()
-    packages_url = \
-        "%s/dists/%s/main/binary-%s/Packages.bz2" % (mirror, distro, arch)
-
+    packages_url = config.get_packages_url(distro)
     logging.debug("Fetching %s" % packages_url)
     f = piupartslib.open_packages_url(packages_url)
     packages_file = piupartslib.packagesdb.PackagesFile(f)
@@ -541,9 +642,10 @@ def create_file(filename, contents):
 def main():
     setup_logging(logging.INFO, None)
 
-    # For supporting multiple architectures and suites, we take a command-line
-    # argument referring to a section in configuration file.  
-    # If no argument is given, the "global" section is assumed.
+    # For supporting multiple architectures and suites, we take command-line
+    # argument(s) referring to section(s) in the configuration file.
+    # If no argument is given, the "sections" entry from the "global" section
+    # is used.
     section_names = []
     global_config = Config(section="global")
     global_config.read(CONFIG_FILE)
@@ -552,16 +654,8 @@ def main():
     else:
         section_names = global_config["sections"].split()
 
-    sections = []
-    for section_name in section_names:
-        section = Section(section_name)
-        section.setup(
-               master_host=global_config["master-host"],
-               master_user=global_config["master-user"],
-               master_directory=global_config["master-directory"],
-               base_tgz_ctrl=[int(global_config["max-tgz-age"]),
-                              int(global_config["min-tgz-retry-delay"])])
-        sections.append(section)
+    sections = [Section(section_name)
+                for section_name in section_names]
 
     while True:
         test_count = 0
@@ -575,9 +669,12 @@ def main():
                     precedence = section.precedence()
 
         if test_count == 0:
-            sleep_time = int(global_config["idle-sleep"])
-            logging.info("Nothing to do, sleeping for %d seconds." % sleep_time)
-            time.sleep( sleep_time )
+            now = time.time()
+            sleep_until = min([now + int(global_config["idle-sleep"])] + [section.sleep_until() for section in sections])
+            if (sleep_until > now):
+                to_sleep = max(60, sleep_until - now)
+                logging.info("Nothing to do, sleeping for %d seconds." % to_sleep)
+                time.sleep(to_sleep)
 
 
 if __name__ == "__main__":
@@ -585,7 +682,7 @@ if __name__ == "__main__":
      main()
   except KeyboardInterrupt:
      print ''
-     print 'Slave interrupted by the user, exiting... manual cleanup still neccessary.'
-     sys.exit(1)  
+     print 'Slave interrupted by the user, exiting...'
+     sys.exit(1)
 
 # vi:set et ts=4 sw=4 :
