@@ -1,17 +1,17 @@
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
-# 
+#
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
 # Free Software Foundation; either version 2 of the License, or (at your
 # option) any later version.
-# 
+#
 # This program is distributed in the hope that it will be useful, but
 # WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 # Public License for more details.
-# 
+#
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc., 
+# this program; if not, write to the Free Software Foundation, Inc.,
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 
 
@@ -27,11 +27,14 @@ Lars Wirzenius <liw@iki.fi>
 import dircache
 import os
 import tempfile
+import time
 import UserDict
-
+import apt_pkg
 
 from piupartslib.dependencyparser import DependencyParser
 
+apt_pkg.init_system()
+
 
 def rfc822_like_header_parse(input):
     headers = []
@@ -55,6 +58,8 @@ class Package(UserDict.UserDict):
             self[name.strip()] = value.strip()
         self._parsed_deps = {}
         self._parsed_alt_deps = {}
+        self._rrdep_count = None
+        self._block_count = None
 
     def _parse_dependencies(self, header_name):
         if header_name in self._parsed_deps:
@@ -85,26 +90,22 @@ class Package(UserDict.UserDict):
         return vlist
 
     # all alternatives - [[package_name...]...]
-    def all_dependencies(self):
+    def all_dependencies(self, header_name=None):
+        headers = ["Depends", "Pre-Depends"]
+        if header_name is not None:
+            headers = [header_name]
         vlist = []
-        for header in ["Depends", "Pre-Depends"]:
+        for header in headers:
             if header in self:
                 vlist += self._parse_alternative_dependencies(header)
         return vlist
 
-    def depends_with_alts(self, header_name):
-        vlist = []
-        if header_name in self:
-            parser = DependencyParser(self[header_name])
-            vlist += parser.get_dependencies()
-        return vlist
-
     def prefer_alt_depends(self, header_name,dep_idx,dep):
         if header_name in self:
             if header_name not in self._parsed_deps:
                   self._parse_dependencies(header_name)
             if self._parsed_deps[header_name][dep_idx]:
-                self._parsed_deps[header_name][dep_idx] = dep.name
+                self._parsed_deps[header_name][dep_idx] = dep
 
     def provides(self):
         vlist = []
@@ -117,6 +118,24 @@ class Package(UserDict.UserDict):
         """Are we testable at all? Required aren't."""
         return self.get("Priority", "") != "required"
 
+    def rrdep_count(self):
+        """Get the recursive dependency count, if it has been calculated"""
+        if self._rrdep_count == None:
+            raise Exception('Reverse dependency count has not been calculated')
+        return(self._rrdep_count)
+
+    def set_rrdep_count(self, val):
+        self._rrdep_count = val
+
+    def block_count(self):
+        """Get the number of packages blocked by this package"""
+        if self._block_count == None:
+            raise Exception('Block count has not been calculated')
+        return(self._block_count)
+
+    def set_block_count(self, val):
+        self._block_count = val
+
     def dump(self, output_file):
         output_file.write("".join(self.headers))
 
@@ -134,6 +153,11 @@ class PackagesFile(UserDict.UserDict):
             if not headers:
                 break
             p = Package(headers)
+            if p["Package"] in self:
+                q = self[p["Package"]]
+                if apt_pkg.version_compare(p["Version"], q["Version"]) <= 0:
+                    # there is already a newer version
+                    continue
             self[p["Package"]] = p
 
 
@@ -168,22 +192,6 @@ class LogDB:
                 return True
         return False
 
-    def any_log_exists(self, package, subdirs):
-        try:
-            cache = self.basename_cache
-        except AttributeError:
-            self.basename_cache = {}
-            cache = self.basename_cache
-        package_name = package["Package"]
-        for subdir in subdirs:
-            for basename in self.listdir(subdir):
-                if basename not in cache:
-                    cache[basename] = basename.split("_", 1)
-                parts = cache[basename]
-                if len(parts) == 2 and parts[0] == package_name:
-                    return True
-        return False
-
     def create(self, subdir, package, version, contents):
         (fd, temp_name) = tempfile.mkstemp(dir=subdir)
         os.close(fd)
@@ -226,26 +234,38 @@ class PackagesDB:
         "dependency-failed-testing",
         "dependency-cannot-be-tested",
         "dependency-does-not-exist",
-        "circular-dependency",
+        "circular-dependency",  # obsolete
         "unknown",
-        "unknown-preferred-alternative",
-        "no-dependency-from-alternatives-exists",
+        "unknown-preferred-alternative",  # obsolete
+        "no-dependency-from-alternatives-exists",  # obsolete
         #"does-not-exist",  # can only happen as query result for a dependency
     ]
 
-    _dep_state_to_state = {
+    _good_states = [
+        "successfully-tested",
+        "essential-required",
+    ]
+
+    _obsolete_states = [
+        "circular-dependency",
+        "unknown-preferred-alternative",
+        "no-dependency-from-alternatives-exists",
+    ]
+
+    _propagate_error_state = {
         "failed-testing": "dependency-failed-testing",
         "cannot-be-tested": "dependency-cannot-be-tested",
-        "waiting-to-be-tested": "waiting-for-dependency-to-be-tested",
-        "waiting-for-dependency-to-be-tested": "waiting-for-dependency-to-be-tested",
         "dependency-failed-testing": "dependency-failed-testing",
         "dependency-cannot-be-tested": "dependency-cannot-be-tested",
-        "dependency-does-not-exist": "dependency-does-not-exist",
-        "unknown-preferred-alternative": "unknown-preferred-alternative",
-        "no-dependency-from-alternatives-exists": "dependency-cannot-be-tested",
+        "dependency-does-not-exist": "dependency-cannot-be-tested",
         "does-not-exist": "dependency-does-not-exist",
     }
 
+    _propagate_waiting_state = {
+        "waiting-to-be-tested": "waiting-for-dependency-to-be-tested",
+        "waiting-for-dependency-to-be-tested": "waiting-for-dependency-to-be-tested",
+    }
+
     def __init__(self, logdb=None, prefix=None):
         self.prefix = prefix
         self._packages_files = []
@@ -255,7 +275,8 @@ class PackagesDB:
         self._in_state = None
         self._package_state = {}
         self.set_subdirs(ok="pass", fail="fail", evil="untestable",
-                         reserved="reserved", morefail=["bugged"])
+                         reserved="reserved", morefail=["bugged", "affected"])
+        self.create_subdirs()
 
     def set_subdirs(self, ok=None, fail=None, evil=None, reserved=None, morefail=None):
         # Prefix all the subdirs with the prefix
@@ -263,6 +284,7 @@ class PackagesDB:
             pformat = self.prefix + "/%s"
         else:
             pformat = "%s"
+        self._submissions = pformat % "submissions.txt"
         if ok:
             self._ok = pformat % ok
         if fail:
@@ -284,10 +306,6 @@ class PackagesDB:
         self._packages_files.append(PackagesFile(input))
         self._packages = None
 
-    def set_known_circular_depends(self, known_circular_depends=[]):
-        self._known_circular_depends = []
-        self._known_circular_depends = list(known_circular_depends)
-
     def _find_all_packages(self):
         if self._packages is None:
             self._packages = {}
@@ -302,7 +320,7 @@ class PackagesDB:
                             self._virtual_packages[provided] = []
                         self._virtual_packages[provided].append(p["Package"])
 
-    def _get_recursive_dependencies(self, package, break_circles=True):
+    def _get_recursive_dependencies(self, package):
         assert self._packages is not None
         deps = []
         more = package.dependencies()
@@ -315,132 +333,130 @@ class PackagesDB:
                     more += self._packages[dep].dependencies()
                 elif dep in self._virtual_packages:
                     more += self._packages[self._virtual_packages[dep][0]].dependencies()
-
-        # Break circular dependencies
-        if break_circles and package["Package"] in deps:
-            deps.remove(package["Package"])
-
         return deps
 
+    def _get_dependency_cycle(self, package_name):
+        deps = []
+        circular = []
+        more = [package_name]
+        while more:
+            dep = more[0]
+            more = more[1:]
+            if dep not in deps:
+                deps.append(dep)
+                if dep in self._packages:
+                    dep_pkg = self._packages[dep]
+                elif dep in self._virtual_packages:
+                    dep_pkg = self._packages[self._virtual_packages[dep][0]]
+                else:
+                    dep_pkg = None
+                if dep_pkg is not None and package_name in self._get_recursive_dependencies(dep_pkg):
+                    circular.append(dep)
+                    more += dep_pkg.dependencies()
+        return circular
+
     def _compute_package_state(self, package):
         if self._logdb.log_exists(package, [self._ok]):
             return "successfully-tested"
         if self._logdb.log_exists(package, [self._fail] + self._morefail):
             return "failed-testing"
-        if self._logdb.any_log_exists(package, [self._evil]):
+        if self._logdb.log_exists(package, [self._evil]):
             return "cannot-be-tested"
         if not package.is_testable():
             return "essential-required"
 
         # First attempt to resolve (still-unresolved) multiple alternative depends
         # Definitely sub-optimal, but improvement over blindly selecting first one
-        # 1) Prefer first alternative = "essential-required", prefer it
-        # 2) If no "essential-required", prefer first alternative = "successfully-tested"
-        # 3) Otherwise, prefer first alternative =  "waiting-to-be-tested" IF NO REMAINING
-        #    are "unknown/fail"
+        # Select the first alternative in the highest of the following states:
+        #   1) "essential-required"
+        #   2) "successfully-tested"
+        #   3) "waiting-to-be-tested" / "waiting-for-dependency-to-be-tested"
+        #   4) "unknown" (retry later)
+        # and update the preferred alternative of that dependency.
+        # If no alternative is in any of these states we retry later ("unknown")
+        # or set "dependency-does-not-exist".
         #
         # Problems:
         #   a) We will test and fail when >=1 "successfully-tested" but another
         #      that failed is selected by apt during test run
-        #   b) False positive "Dependency failed/cannot be tested"; however
-        #       more accurately "waiting-for-dependency-to-be-tested"
+        #   b) We may report a status of "waiting-for-dependency-to-be-tested"
+        #      instead of "waiting-to-be-tested" depending on the order the
+        #      package states get resolved.
 
-        state = None
         for header in ["Depends", "Pre-Depends"]:
-            alt_deps=package.depends_with_alts(header)
+            alt_deps = package.all_dependencies(header)
             for d in range(len(alt_deps)):
                 if len(alt_deps[d]) > 1:
                     alt_found = 0
-                    alt_fails = 0
-                    alt_unknowns = 0
-                    alt_state = None
-                    prefer_alt_score = 0
-                    prefer_alt_idx = 0
+                    prefer_alt_score = -1
                     prefer_alt = None
                     for alternative in alt_deps[d]:
-                        dep = alternative.name
-                        altdep_state = self.get_package_state(dep)
+                        altdep_state = self.get_package_state(alternative)
                         if altdep_state != "does-not-exist":
                             alt_found += 1
                             if prefer_alt_score < 3 and altdep_state == "essential-required":
                                 prefer_alt = alternative
-                                prefer_alt_idx = d
                                 prefer_alt_score = 3
                             elif prefer_alt_score < 2 and altdep_state == "successfully-tested":
                                 prefer_alt = alternative
-                                prefer_alt_idx = d
                                 prefer_alt_score = 2
                             elif prefer_alt_score < 1 and \
                                  altdep_state in ["waiting-to-be-tested", "waiting-for-dependency-to-be-tested"]:
                                 prefer_alt = alternative
-                                prefer_alt_idx = d
                                 prefer_alt_score = 1
-                            elif altdep_state == "unknown":
-                                alt_unknowns += 1
-                            else:
-                                alt_fails += 1
-                                if alt_state is None:
-                                    alt_state = altdep_state
+                            elif prefer_alt_score < 0 and altdep_state == "unknown":
+                                prefer_alt = alternative
+                                prefer_alt_score = 0
+                    if alt_found == 0:
+                        return "dependency-does-not-exist"
+                    if prefer_alt_score >= 0:
+                        package.prefer_alt_depends(header, d, prefer_alt)
 
-                    if prefer_alt_score >= 2:
-                        package.prefer_alt_depends(header,prefer_alt_idx,prefer_alt)
-                    elif prefer_alt_score == 1 and ((alt_unknowns + alt_fails) == 0):
-                        package.prefer_alt_depends(header,prefer_alt_idx,prefer_alt)
-                    elif alt_found == 0:
-                        return "no-dependency-from-alternatives-exists"
-                    else:
-                        if alt_state is not None and alt_unknowns == 0:
-                            state = alt_state
-                        elif state is None:
-                            state = "unknown-preferred-alternative"
+        deps = package.dependencies()
 
-        if state is not None:
-             return state
-
-        for dep in package.dependencies():
+        for dep in deps:
             dep_state = self.get_package_state(dep)
-            if dep_state in self._dep_state_to_state:
-                return self._dep_state_to_state[dep_state]
+            if dep_state in self._propagate_error_state:
+                return self._propagate_error_state[dep_state]
 
-        state = "waiting-to-be-tested"
-        for dep in package.dependencies():
+        testable = True
+        for dep in deps:
             dep_state = self.get_package_state(dep)
-            if dep_state not in \
-                    ["successfully-tested", "essential-required"]:
-                state = "unknown"
+            if dep_state not in self._good_states:
+                testable = False
                 break
-        if state == "waiting-to-be-tested":
-            return state
-
-        deps = self._get_recursive_dependencies(package, break_circles=False)
-        # ignore those packages:
-        for pkg in self._known_circular_depends:
-            if pkg in deps:
-                deps.remove(pkg)
-        if package["Package"] in deps:
-            return "circular-dependency"  # actually, it's an unknown circular-dependency
+        if testable:
+            return "waiting-to-be-tested"
 
         # treat circular-dependencies as testable (for the part of the circle)
-        state = "unknown" 
-        if package["Package"] in self._known_circular_depends:
-            for dep in package.dependencies():
+        circular_deps = self._get_dependency_cycle(package["Package"])
+        if package["Package"] in circular_deps:
+            testable = True
+            for dep in deps:
                 dep_state = self.get_package_state(dep)
-                if dep not in self._known_circular_depends and dep_state not in \
-                        ["successfully-tested", "essential-required"]:
-                    state = "unknown"
+                if dep in circular_deps:
+                    # allow any non-error dep_state on the cycle for testing
+                    # (error states are handled by the error propagation above)
+                    pass
+                elif dep_state not in self._good_states:
+                    # non-circular deps must have passed before testing circular deps
+                    testable = False
                     break
-                if dep in self._known_circular_depends and dep_state not in \
-                        ["failed-testing", "dependency-failed-testing"]:
-                    state = "waiting-to-be-tested"
-                    continue
-        return state
+            if testable:
+                return "waiting-to-be-tested"
+
+        for dep in deps:
+            dep_state = self.get_package_state(dep)
+            if dep_state in self._propagate_waiting_state:
+                return self._propagate_waiting_state[dep_state]
+
+        return "unknown"
 
     def _compute_package_states(self):
         if self._in_state is not None:
             return
 
         todo = []
-        unpreferred_alt = []
 
         self._find_all_packages()
         package_names = self._packages.keys()
@@ -456,17 +472,13 @@ class PackagesDB:
         while package_names:
             todo = []
             done = []
-            unpreferred_alt = []
             for package_name in package_names:
                 package = self._packages[package_name]
-                if self._package_state[package_name] in \
-                   [ "unknown", "unknown-preferred-alternative" ]:
+                if self._package_state[package_name] == "unknown":
                     state = self._compute_package_state(package)
                     assert state in self._states
                     if state == "unknown":
                         todo.append(package_name)
-                    elif state == "unknown-preferred-alternative":
-                        unpreferred_alt.append(package_name)
                     else:
                         self._in_state[state].append(package_name)
                         self._package_state[package_name] = state
@@ -476,12 +488,8 @@ class PackagesDB:
                 # to do anything the next time either.
                 break
             package_names = todo
-            package_names.extend(unpreferred_alt)
 
         self._in_state["unknown"] = todo
-        self._in_state["unknown-preferred-alternative"] = unpreferred_alt
-        for package_name in unpreferred_alt:
-            self._package_state[package_name] = "unknown-preferred-alternative"
 
         for state in self._states:
             self._in_state[state].sort()
@@ -489,6 +497,12 @@ class PackagesDB:
     def get_states(self):
         return self._states
 
+    def get_active_states(self):
+        return [x for x in self._states if not x in self._obsolete_states]
+
+    def get_error_states(self):
+        return [x for x in self._propagate_error_state.keys() if x in self._states]
+
     def get_pkg_names_in_state(self, state):
         self._compute_package_states()
         return set(self._in_state[state])
@@ -515,7 +529,7 @@ class PackagesDB:
           # don't have a Source header, so let's try:
           try:
             _source = self._packages[package_name][header]
-            # for binNMU the Source header in Packages files holds the version 
+            # for binNMU the Source header in Packages files holds the version
             # too, so we need to chop it of:
             if " " in _source:
               source, version = _source.split(" ")
@@ -561,6 +575,10 @@ class PackagesDB:
         if "/" in str:
             raise Exception("'/' in (partial) filename: %s" % str)
 
+    def _record_submission(self, category, package, version):
+        with open(self._submissions, "a") as submissions:
+            submissions.write("%d %s %s %s\n" % (time.time(), category, package, version))
+
     def unreserve_package(self, package, version):
         self._check_for_acceptability_as_filename(package)
         self._check_for_acceptability_as_filename(version)
@@ -571,6 +589,7 @@ class PackagesDB:
         self._check_for_acceptability_as_filename(version)
         if self._logdb.create(self._ok, package, version, log):
             self._logdb.remove(self._reserved, package, version)
+            self._record_submission("pass", package, version)
         else:
             raise Exception("Log file exists already: %s (%s)" %
                                 (package, version))
@@ -580,6 +599,7 @@ class PackagesDB:
         self._check_for_acceptability_as_filename(version)
         if self._logdb.create(self._fail, package, version, log):
             self._logdb.remove(self._reserved, package, version)
+            self._record_submission("fail", package, version)
         else:
             raise Exception("Log file exists already: %s (%s)" %
                                 (package, version))
@@ -589,8 +609,56 @@ class PackagesDB:
         self._check_for_acceptability_as_filename(version)
         if self._logdb.create(self._evil, package, version, log):
             self._logdb.remove(self._reserved, package, version)
+            self._record_submission("untestable", package, version)
         else:
             raise Exception("Log file exists already: %s (%s)" %
                                 (package, version))
 
+    def calc_rrdep_counts(self):
+        """Calculate recursive reverse dependency counts for Packages"""
+
+        self._find_all_packages()       # populate _packages
+        self._compute_package_states()  # populate _package_state
+        error_states = self.get_error_states()
+
+        # create a reverse dependency dictionary.
+        # entries consist of a one-level list of reverse dependency package names,
+        # by package name
+        rdeps = {}
+        for pkg_name in self._packages.keys():
+            # use the Packages dependencies() method for a conservative count
+            for dep in self._packages[pkg_name].dependencies():
+                if dep in rdeps:
+                    rdeps[dep].append( pkg_name )
+                else:
+                    rdeps[dep] = [pkg_name]
+
+        def recurse_rdeps( pkg_name, rdeps, rrdep_dict ):
+            """ Recurse through the reverse dep arrays to determine the recursive
+                dependency count for a package. rrdep_dict.keys() contains the
+                accumulation of rdeps encountered"""
+
+            # are there any rdeps for this package?
+            if pkg_name in rdeps:
+                for rdep in rdeps[pkg_name]:
+                    # break circular dependency loops
+                    if not rdep in rrdep_dict:
+                        rrdep_dict[rdep] = 1
+                        rrdep_dict = recurse_rdeps( rdep, rdeps, rrdep_dict )
+
+            return rrdep_dict
+
+        # calculate all of the rrdeps and block counts
+        for pkg_name in self._packages.keys():
+            rrdep_list = recurse_rdeps( pkg_name, rdeps, {} ).keys()
+            self._packages[pkg_name].set_rrdep_count( len(rrdep_list) )
+
+            if self._package_state[pkg_name] in error_states:
+                block_list = [x for x in rrdep_list
+                              if self._package_state[x] in error_states]
+            else:
+                block_list = []
+            self._packages[pkg_name].set_block_count( len(block_list) )
+
+
 # vi:set et ts=4 sw=4 :
